# -*- coding: utf-8 -*-

from functools import wraps
from flask import jsonify, request
from marshmallow import ValidationError, Schema
from .exceptions import IncompatibleSchema
from furl import furl

__DEFAULT_STRICT_SLASHES = False


def _validate_schema(obj):
    """Ensures the passed schema instance is compatible

    :param obj: object to validate
    :return: obj
    :raises:
        - IncompatibleSchema if the passed schema is of an incompatible type
    """

    if obj is not None and not isinstance(obj, Schema):
        raise IncompatibleSchema('Schema must be of type {0}'.format(Schema))

    return obj


def route(bp, *args, strict_slashes=__DEFAULT_STRICT_SLASHES,
          body_schema=None, query_schema=None, marshal_with=None, **kwargs):

    """Journey route decorator

    Enables simple serialization, deserialization and validation of Flask routes with the help of Marshmallow.

    If deserialization (body_schema or query_schema) was passed to the decorator, the corresponding
    :class`marshmallow.Schema` object gets passed to the decorated function.

    __query - kwarg if `query_schema` was passed
    __body - kwarg if `body_schema` was passed


    :param bp: :class:`flask.Blueprint` object
    :param strict_slashes: Enable / disable strict slashes (default False)
    :param body_schema: Deserialize JSON body with this schema
    :param query_schema: Deserialize Query string with this schema
    :param marshal_with: Serialize the output with this schema
    :param args: args to pass along to `Blueprint.route`
    :param kwargs: kwargs to pass along to `Blueprint.route`
    """

    kwargs['strict_slashes'] = strict_slashes
    body_schema = _validate_schema(body_schema)
    query_schema = _validate_schema(query_schema)
    marshal_with = _validate_schema(marshal_with)

    def decorator(f):
        @bp.route(*args, **kwargs)
        @wraps(f)
        def wrapper(*inner_args, **inner_kwargs):
            try:
                if query_schema:
                    url = furl(request.url)
                    qs_data, qs_errors = query_schema.load(data=url.args)
                    inner_kwargs['__query'] = {'data': qs_data, 'errors': qs_errors}

                if body_schema:
                    json_data = request.get_json()
                    bs_data, bs_errors = body_schema.load(data=json_data)
                    inner_kwargs['__body'] = {'data': bs_data, 'errors': bs_errors}

            except ValidationError as err:
                return jsonify(err.messages), 422

            if marshal_with:
                data = marshal_with.dump(f(*inner_args, **inner_kwargs))
                return jsonify(data[0])

            return f(*inner_args, **inner_kwargs)

        return f

    return decorator



# -*- coding: utf-8 -*-

"""
    Flask-Journey
    ~~~~~~~~~~~~
    Simple Blueprint / route manager for Flask

"""

__author__ = "Robert Wikman <rbw@vault13.org>"
__version__ = "0.1.0"

from .exceptions import InvalidBasePath, IncompatibleRoute


class Route(object):
    """Creates a new route on the path specified

    :param path: route path
    """

    def __init__(self, path='/', description=''):
        self.path = self.sanitize_path(path)
        self.description = description
        self.blueprints = []

    @staticmethod
    def sanitize_path(path):
        """Performs sanitation of the route path after validating

        :param path: path to sanitize
        :return: sanitized path
        """

        if path[:1] != '/':
            raise InvalidBasePath('The manager base path must start with a slash')

        return path.rstrip('/')

    def attach_bp(self, bp, description=''):
        """Attaches a blueprint

        :param bp: :class:`flask.Blueprint` object
        :param description: Optional description string
        """

        self.blueprints.append((bp, description))


class Journey(object):
    """Central controller class.
    Registers blueprints and exposes `routes` property, which lists all routes added with Journey.
    :param app: App to pass directly to Journey
    """

    def __init__(self, app=None):
        self._app = None
        self._routes = []

        if app is not None:
            self.init_app(app)

    def init_app(self, app):
        """Initializes Journey extension
        :param app: App passed from constructor or directly to init_app
        """

        self._app = app

    @property
    def routes_detailed(self):
        """Returns a detailed list of routes added with Journey

        :return: List of blueprint routes
        """

        return self._routes

    @property
    def routes_simple(self):
        """Returns a list of tuples containing endpoint-route pairs

        :return: List of route pairs
        """

        routes = []

        for route in self._routes:
            base_path = route['base_path']
            for blueprint in route['blueprints']:
                bp_path = blueprint['path']
                for child in blueprint['routes']:
                    routes.append(
                        (
                            child['endpoint'],
                            base_path + bp_path + child['path'],
                            child['methods']
                        )
                    )

        return routes

    def register_route(self, route):
        """Registers blueprints attached to the :class:`flask_journey.Route` object passed

        :param route: :class:`flask_journey.Route` object
        :raises:
            - IncompatibleRoute if the route is not of type `Route`
        """

        if not isinstance(route, Route):
            raise IncompatibleRoute('Manager object passed to register_router must be of type {0}'.format(Route))

        registered_route = {
            'base_path': route.path,
            'description': route.description,
            'blueprints': []
        }

        for (bp, description) in route.blueprints:
            # Get route name (from url_prefix or name attr) and prepend with a slash
            child_path = self.get_child_path(bp)

            # Create full path to the BP route
            base_path = route.path + child_path

            # Register the BP
            blueprint = self._register_blueprint(bp, base_path, child_path, description)

            # Add routes for this blueprint
            blueprint['routes'] = self._get_blueprint_routes(base_path)

            # Finally, attach the blueprints to its parent
            registered_route['blueprints'].append(blueprint)

        self._routes.append(registered_route)

    def _register_blueprint(self, bp, base_path, child_path, description):
        """Register and return info about the registered blueprint

        :param bp: :class:`flask.Blueprint` object
        :param base_path: the URL prefix of this blueprint
        :param child_path: blueprint relative to the router path
        :return: Dict with info about the blueprint
        """

        self._app.register_blueprint(bp, url_prefix=base_path)
        return {
            'name': bp.name,
            'path': child_path,
            'import_name': bp.import_name,
            'description': description,
            'routes': []
        }

    @staticmethod
    def get_child_path(bp):
        """Strips leading slashes from url_prefix, if it has been set, and returns. If not, return `Blueprint.name`.

        :param bp: :class:`flask.Blueprint` object
        :return: blueprint name
        """

        return bp.url_prefix or '/' + bp.name

    def _get_blueprint_routes(self, base_path):
        """Returns detailed information about registered blueprint routes matching the `Route` path

        :param base_path: Base path to return detailed route info for
        :return: List of route detail dicts
        """

        routes = []

        for child in self._app.url_map.iter_rules():
            if child.rule.startswith(base_path):
                relative_path = child.rule[len(base_path):]
                routes.append({
                    'path': relative_path,
                    'endpoint': child.endpoint,
                    'methods': list(child.methods)
                })

        return routes
